#include <cinttypes>
#include <cstdio>
#include <windows.h>
#include <d3d11_1.h>

#define MODE_NV12 0
#define MODE_R8_R8G8 1
#define MODE_RGBA8_PACKED 2 

#define COPY_MODE MODE_NV12

HINSTANCE                           g_hInst = nullptr;
HWND                                g_hWnd = nullptr;
D3D_DRIVER_TYPE                     g_driverType = D3D_DRIVER_TYPE_NULL;
D3D_FEATURE_LEVEL                   g_featureLevel = D3D_FEATURE_LEVEL_11_0;
ID3D11Device*                       g_pd3dDevice = nullptr;
ID3D11Device1*                      g_pd3dDevice1 = nullptr;
ID3D11DeviceContext*                g_pImmediateContext = nullptr;
ID3D11DeviceContext1*               g_pImmediateContext1 = nullptr;
IDXGISwapChain*                     g_pSwapChain = nullptr;
IDXGISwapChain1*                    g_pSwapChain1 = nullptr;
ID3D11RenderTargetView*             g_pRenderTargetView = nullptr;
ID3D11Texture2D*                    g_pYUV = nullptr;
ID3D11RenderTargetView*             g_pYUVTargetView = nullptr;
ID3D11ShaderResourceView*           g_pYUVShaderView = nullptr;
ID3D11Texture2D*                    g_pNV12 = nullptr;
ID3D11Texture2D*                    g_pStageNV12[2]{};
ID3D11Texture2D*                    g_pR8 = nullptr;
ID3D11Texture2D*                    g_pStageR8[2]{};
ID3D11Texture2D*                    g_pR8G8 = nullptr;
ID3D11Texture2D*                    g_pStageR8G8[2]{};
ID3D11Texture2D*                    g_pR8G8B8A8 = nullptr;
ID3D11Texture2D*                    g_pStageR8G8B8A8[2]{};
ID3D11Query*                        g_queryDisjoint[2]{};
ID3D11Query*                        g_queryStart[2]{};
ID3D11Query*                        g_queryEnd[2]{};
uint32_t                            g_frameIndex;

HRESULT InitWindow(HINSTANCE hInstance, int nCmdShow);
HRESULT InitDevice();
void CleanupDevice();
LRESULT CALLBACK    WndProc(HWND, UINT, WPARAM, LPARAM);
void Render();

int WINAPI wWinMain(_In_ HINSTANCE hInstance, _In_opt_ HINSTANCE hPrevInstance, _In_ LPWSTR lpCmdLine, _In_ int nCmdShow)
{
	UNREFERENCED_PARAMETER(hPrevInstance);
	UNREFERENCED_PARAMETER(lpCmdLine);

	if (FAILED(InitWindow(hInstance, nCmdShow)))
		return 0;

	if (FAILED(InitDevice()))
	{
		CleanupDevice();
		return 0;
	}

	// Main message loop
	MSG msg = { 0 };
	while (WM_QUIT != msg.message)
	{
		if (PeekMessage(&msg, nullptr, 0, 0, PM_REMOVE))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		else
		{
			Render();
		}
	}

	CleanupDevice();

	return (int)msg.wParam;
}

HRESULT InitWindow(HINSTANCE hInstance, int nCmdShow)
{
	// Register class
	WNDCLASSEX wcex;
	wcex.cbSize = sizeof(WNDCLASSEX);
	wcex.style = CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc = WndProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hInstance;
	wcex.hIcon = nullptr;
	wcex.hCursor = LoadCursor(nullptr, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
	wcex.lpszMenuName = nullptr;
	wcex.lpszClassName = L"CopySpeedsWindowClass";
	wcex.hIconSm = nullptr;
	if (!RegisterClassEx(&wcex))
		return E_FAIL;

	// Create window
	g_hInst = hInstance;
	RECT rc = { 0, 0, 1920, 1080 };
	AdjustWindowRect(&rc, WS_OVERLAPPEDWINDOW, FALSE);
	g_hWnd = CreateWindow(L"CopySpeedsWindowClass", L"CopySpeeds",
		WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX,
		CW_USEDEFAULT, CW_USEDEFAULT, rc.right - rc.left, rc.bottom - rc.top, nullptr, nullptr, hInstance,
		nullptr);
	if (!g_hWnd)
		return E_FAIL;

	ShowWindow(g_hWnd, nCmdShow);

	return S_OK;
}

HRESULT InitDevice()
{
	HRESULT hr = S_OK;

	RECT rc;
	GetClientRect(g_hWnd, &rc);
	UINT width = rc.right - rc.left;
	UINT height = rc.bottom - rc.top;

	UINT createDeviceFlags = 0;
#ifdef _DEBUG
	createDeviceFlags |= D3D11_CREATE_DEVICE_DEBUG;
#endif

	D3D_DRIVER_TYPE driverTypes[] =
	{
		D3D_DRIVER_TYPE_HARDWARE,
		D3D_DRIVER_TYPE_WARP,
		D3D_DRIVER_TYPE_REFERENCE,
	};
	UINT numDriverTypes = ARRAYSIZE(driverTypes);

	D3D_FEATURE_LEVEL featureLevels[] =
	{
		D3D_FEATURE_LEVEL_11_1,
		D3D_FEATURE_LEVEL_11_0,
		D3D_FEATURE_LEVEL_10_1,
		D3D_FEATURE_LEVEL_10_0,
	};
	UINT numFeatureLevels = ARRAYSIZE(featureLevels);

	for (UINT driverTypeIndex = 0; driverTypeIndex < numDriverTypes; driverTypeIndex++)
	{
		g_driverType = driverTypes[driverTypeIndex];
		hr = D3D11CreateDevice(nullptr, g_driverType, nullptr, createDeviceFlags, featureLevels, numFeatureLevels,
			D3D11_SDK_VERSION, &g_pd3dDevice, &g_featureLevel, &g_pImmediateContext);

		if (hr == E_INVALIDARG)
		{
			// DirectX 11.0 platforms will not recognize D3D_FEATURE_LEVEL_11_1 so we need to retry without it
			hr = D3D11CreateDevice(nullptr, g_driverType, nullptr, createDeviceFlags, &featureLevels[1], numFeatureLevels - 1,
				D3D11_SDK_VERSION, &g_pd3dDevice, &g_featureLevel, &g_pImmediateContext);
		}

		if (SUCCEEDED(hr))
			break;
	}
	if (FAILED(hr))
		return hr;

	// Obtain DXGI factory from device (since we used nullptr for pAdapter above)
	IDXGIFactory1* dxgiFactory = nullptr;
	{
		IDXGIDevice* dxgiDevice = nullptr;
		hr = g_pd3dDevice->QueryInterface(__uuidof(IDXGIDevice), reinterpret_cast<void**>(&dxgiDevice));
		if (SUCCEEDED(hr))
		{
			IDXGIAdapter* adapter = nullptr;
			hr = dxgiDevice->GetAdapter(&adapter);
			if (SUCCEEDED(hr))
			{
				DXGI_ADAPTER_DESC desc;
				adapter->GetDesc(&desc);

				hr = adapter->GetParent(__uuidof(IDXGIFactory1), reinterpret_cast<void**>(&dxgiFactory));
				adapter->Release();
			}
			dxgiDevice->Release();
		}
	}
	if (FAILED(hr))
		return hr;

	// Create swap chain
	IDXGIFactory2* dxgiFactory2 = nullptr;
	hr = dxgiFactory->QueryInterface(__uuidof(IDXGIFactory2), reinterpret_cast<void**>(&dxgiFactory2));
	if (dxgiFactory2)
	{
		// DirectX 11.1 or later
		hr = g_pd3dDevice->QueryInterface(__uuidof(ID3D11Device1), reinterpret_cast<void**>(&g_pd3dDevice1));
		if (SUCCEEDED(hr))
		{
			(void)g_pImmediateContext->QueryInterface(__uuidof(ID3D11DeviceContext1), reinterpret_cast<void**>(&g_pImmediateContext1));
		}

		DXGI_SWAP_CHAIN_DESC1 sd = {};
		sd.Width = width;
		sd.Height = height;
		sd.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
		sd.SampleDesc.Count = 1;
		sd.SampleDesc.Quality = 0;
		sd.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;
		sd.BufferCount = 2;
		sd.SwapEffect = DXGI_SWAP_EFFECT_FLIP_DISCARD;

		hr = dxgiFactory2->CreateSwapChainForHwnd(g_pd3dDevice, g_hWnd, &sd, nullptr, nullptr, &g_pSwapChain1);
		if (SUCCEEDED(hr))
		{
			hr = g_pSwapChain1->QueryInterface(__uuidof(IDXGISwapChain), reinterpret_cast<void**>(&g_pSwapChain));
		}

		dxgiFactory2->Release();
	}
	else
	{
		// DirectX 11.0 systems
		DXGI_SWAP_CHAIN_DESC sd = {};
		sd.BufferCount = 1;
		sd.BufferDesc.Width = width;
		sd.BufferDesc.Height = height;
		sd.BufferDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
		sd.BufferDesc.RefreshRate.Numerator = 60;
		sd.BufferDesc.RefreshRate.Denominator = 1;
		sd.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;
		sd.OutputWindow = g_hWnd;
		sd.SampleDesc.Count = 1;
		sd.SampleDesc.Quality = 0;
		sd.Windowed = TRUE;

		hr = dxgiFactory->CreateSwapChain(g_pd3dDevice, &sd, &g_pSwapChain);
	}

	dxgiFactory->MakeWindowAssociation(g_hWnd, DXGI_MWA_NO_ALT_ENTER);

	dxgiFactory->Release();

	if (FAILED(hr))
		return hr;

	for (int i = 0; i < 2; ++i)
	{
		D3D11_QUERY_DESC desc{};
		desc.Query = D3D11_QUERY_TIMESTAMP_DISJOINT;
		g_pd3dDevice->CreateQuery(&desc, &g_queryDisjoint[i]);

		desc.Query = D3D11_QUERY_TIMESTAMP;
		g_pd3dDevice->CreateQuery(&desc, &g_queryStart[i]);
		g_pd3dDevice->CreateQuery(&desc, &g_queryEnd[i]);
	}

	// Create a render target view
	ID3D11Texture2D* pBackBuffer = nullptr;
	hr = g_pSwapChain->GetBuffer(0, __uuidof(ID3D11Texture2D), reinterpret_cast<void**>(&pBackBuffer));
	if (FAILED(hr))
		return hr;

	hr = g_pd3dDevice->CreateRenderTargetView(pBackBuffer, nullptr, &g_pRenderTargetView);
	pBackBuffer->Release();
	if (FAILED(hr))
		return hr;

	D3D11_TEXTURE2D_DESC descNV12 = {};
	descNV12.Width = 1920;
	descNV12.Height = 1080;
	descNV12.MipLevels = 1;
	descNV12.ArraySize = 1;
	descNV12.Format = DXGI_FORMAT_NV12;
	descNV12.SampleDesc.Count = 1;
	descNV12.SampleDesc.Quality = 0;
	descNV12.Usage = D3D11_USAGE_DEFAULT;
	descNV12.BindFlags = D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_RENDER_TARGET;
	descNV12.CPUAccessFlags = 0;
	descNV12.MiscFlags = 0;
	hr = g_pd3dDevice->CreateTexture2D(&descNV12, nullptr, &g_pNV12);
	if (FAILED(hr))
		return hr;
	descNV12.Usage = D3D11_USAGE_STAGING;
	descNV12.BindFlags = 0;
	descNV12.CPUAccessFlags = D3D11_CPU_ACCESS_READ;
	hr = g_pd3dDevice->CreateTexture2D(&descNV12, nullptr, &g_pStageNV12[0]);
	if (FAILED(hr))
		return hr;
	hr = g_pd3dDevice->CreateTexture2D(&descNV12, nullptr, &g_pStageNV12[1]);
	if (FAILED(hr))
		return hr;

	D3D11_TEXTURE2D_DESC descR8 = {};
	descR8.Width = 1920;
	descR8.Height = 1080;
	descR8.MipLevels = 1;
	descR8.ArraySize = 1;
	descR8.Format = DXGI_FORMAT_R8_UNORM;
	descR8.SampleDesc.Count = 1;
	descR8.SampleDesc.Quality = 0;
	descR8.Usage = D3D11_USAGE_DEFAULT;
	descR8.BindFlags = D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_RENDER_TARGET;
	descR8.CPUAccessFlags = 0;
	descR8.MiscFlags = 0;
	hr = g_pd3dDevice->CreateTexture2D(&descR8, nullptr, &g_pR8);
	if (FAILED(hr))
		return hr;
	descR8.Usage = D3D11_USAGE_STAGING;
	descR8.BindFlags = 0;
	descR8.CPUAccessFlags = D3D11_CPU_ACCESS_READ;
	hr = g_pd3dDevice->CreateTexture2D(&descR8, nullptr, &g_pStageR8[0]);
	if (FAILED(hr))
		return hr;
	hr = g_pd3dDevice->CreateTexture2D(&descR8, nullptr, &g_pStageR8[1]);
	if (FAILED(hr))
		return hr;

	D3D11_TEXTURE2D_DESC descR8G8 = {};
	descR8G8.Width = 960;
	descR8G8.Height = 540;
	descR8G8.MipLevels = 1;
	descR8G8.ArraySize = 1;
	descR8G8.Format = DXGI_FORMAT_R8G8_UNORM;
	descR8G8.SampleDesc.Count = 1;
	descR8G8.SampleDesc.Quality = 0;
	descR8G8.Usage = D3D11_USAGE_DEFAULT;
	descR8G8.BindFlags = D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_RENDER_TARGET;
	descR8G8.CPUAccessFlags = 0;
	descR8G8.MiscFlags = 0;
	hr = g_pd3dDevice->CreateTexture2D(&descR8G8, nullptr, &g_pR8G8);
	if (FAILED(hr))
		return hr;
	descR8G8.Usage = D3D11_USAGE_STAGING;
	descR8G8.BindFlags = 0;
	descR8G8.CPUAccessFlags = D3D11_CPU_ACCESS_READ;
	hr = g_pd3dDevice->CreateTexture2D(&descR8G8, nullptr, &g_pStageR8G8[0]);
	if (FAILED(hr))
		return hr;
	hr = g_pd3dDevice->CreateTexture2D(&descR8G8, nullptr, &g_pStageR8G8[1]);
	if (FAILED(hr))
		return hr;

	D3D11_TEXTURE2D_DESC descR8G8B8A8 = {};
	descR8G8B8A8.Width = 1920;
	descR8G8B8A8.Height = 406;
	descR8G8B8A8.MipLevels = 1;
	descR8G8B8A8.ArraySize = 1;
	descR8G8B8A8.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
	descR8G8B8A8.SampleDesc.Count = 1;
	descR8G8B8A8.SampleDesc.Quality = 0;
	descR8G8B8A8.Usage = D3D11_USAGE_DEFAULT;
	descR8G8B8A8.BindFlags = D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_RENDER_TARGET;
	descR8G8B8A8.CPUAccessFlags = 0;
	descR8G8B8A8.MiscFlags = 0;
	hr = g_pd3dDevice->CreateTexture2D(&descR8G8B8A8, nullptr, &g_pR8G8B8A8);
	if (FAILED(hr))
		return hr;
	descR8G8B8A8.Usage = D3D11_USAGE_STAGING;
	descR8G8B8A8.BindFlags = 0;
	descR8G8B8A8.CPUAccessFlags = D3D11_CPU_ACCESS_READ;
	hr = g_pd3dDevice->CreateTexture2D(&descR8G8B8A8, nullptr, &g_pStageR8G8B8A8[0]);
	if (FAILED(hr))
		return hr;
	hr = g_pd3dDevice->CreateTexture2D(&descR8G8B8A8, nullptr, &g_pStageR8G8B8A8[1]);
	if (FAILED(hr))
		return hr;

	return S_OK;
}

void CleanupDevice()
{
	if (g_pImmediateContext) g_pImmediateContext->ClearState();
	if (g_pRenderTargetView) g_pRenderTargetView->Release();
	if (g_pNV12) g_pNV12->Release();
	for (int i = 0; i < 2; ++i)
	{
		if (g_pStageNV12[i]) g_pStageNV12[i]->Release();
		if (g_pStageR8[i]) g_pStageR8[i]->Release();
		if (g_pStageR8G8[i]) g_pStageR8G8[i]->Release();
		if (g_pStageR8G8B8A8[i]) g_pStageR8G8B8A8[i]->Release();
		if (g_queryDisjoint[i]) g_queryDisjoint[i]->Release();
		if (g_queryStart[i]) g_queryStart[i]->Release();
		if (g_queryEnd[i]) g_queryEnd[i]->Release();
	}
	if (g_pR8) g_pR8->Release();
	if (g_pR8G8) g_pR8G8->Release();
	if (g_pR8G8B8A8) g_pR8G8B8A8->Release();
	if (g_pSwapChain1) g_pSwapChain1->Release();
	if (g_pSwapChain) g_pSwapChain->Release();
	if (g_pImmediateContext1) g_pImmediateContext1->Release();
	if (g_pImmediateContext) g_pImmediateContext->Release();
	if (g_pd3dDevice1) g_pd3dDevice1->Release();
	if (g_pd3dDevice) g_pd3dDevice->Release();
}


LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT ps;
	HDC hdc;

	switch (message)
	{
	case WM_PAINT:
		hdc = BeginPaint(hWnd, &ps);
		EndPaint(hWnd, &ps);
		break;

	case WM_DESTROY:
		PostQuitMessage(0);
		break;

	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}

	return 0;
}

double min_us = 100000.0;

void Render()
{
	g_pImmediateContext->Begin(g_queryDisjoint[g_frameIndex]);
	g_pImmediateContext->End(g_queryStart[g_frameIndex]);

	const uint32_t nextIndex = 1 - g_frameIndex;

#if COPY_MODE == MODE_NV12
	g_pImmediateContext->CopyResource(g_pStageNV12[g_frameIndex], g_pNV12);
#elif COPY_MODE == MODE_R8_R8G8
	g_pImmediateContext->CopyResource(g_pStageR8[g_frameIndex], g_pR8);
	g_pImmediateContext->CopyResource(g_pStageR8G8[g_frameIndex], g_pR8G8);
#elif COPY_MODE == MODE_RGBA8_PACKED
	g_pImmediateContext->CopyResource(g_pStageR8G8B8A8[g_frameIndex], g_pR8G8B8A8);
#endif

	g_pImmediateContext->End(g_queryEnd[g_frameIndex]);
	g_pImmediateContext->End(g_queryDisjoint[g_frameIndex]);

	uint64_t start, end;
	HRESULT hr_start, hr_end, hr_disjoint;
	D3D11_QUERY_DATA_TIMESTAMP_DISJOINT disjoint;
	do
	{
		hr_start = g_pImmediateContext->GetData(g_queryStart[nextIndex], &start, sizeof(start), 0);
	} while (hr_start == S_FALSE);
	do
	{
		hr_end = g_pImmediateContext->GetData(g_queryEnd[nextIndex], &end, sizeof(end), 0);
	} while (hr_end == S_FALSE);
	do
	{
		hr_disjoint = g_pImmediateContext->GetData(g_queryDisjoint[nextIndex], &disjoint, sizeof(disjoint), 0);
	} while (hr_disjoint == S_FALSE);

	if (SUCCEEDED(hr_start) && SUCCEEDED(hr_end) && SUCCEEDED(hr_disjoint))
	{
		if (!disjoint.Disjoint)
		{
			double us = 1000000.0 * (double)(end - start) / (double)disjoint.Frequency;
			if (us < min_us)
			{
				min_us = us;
				WCHAR buffer[64];
				swprintf(buffer, _countof(buffer), L"Copy Duration: %f us\n", min_us);
				OutputDebugString(buffer);
			}
		}
	}

	//
	// Present our back buffer to our front buffer
	//
	g_pSwapChain->Present(1, 0);

	g_frameIndex = nextIndex;
}
